from __future__ import annotations
from typing import List, Union

from ..logic_variables import euc, PyValue, n_Vars, unify, unify_pairs, unify_sequences, Var
from ..sequence_options.super_sequence import SuperSequence


class PySequence(SuperSequence):
  """
  Python treats lists and tuples as essentially the same. This is the common core.
  The self.args are the list/tuple elements. Their length is fixed. (This disallows
  appending elements to a list or extending a list.)
  """
  def __init__(self, pyType, initialElements: Union[list, tuple]):
    super().__init__( (pyType, *initialElements) )

  def __add__(self, Other: Union[PySequence, Var]) -> PySequence:
    Other_EoT = Other.unification_chain_end()
    # If not, can't append.
    assert isinstance(Other_EoT, PySequence)
    Result = Var()
    for _ in append(self, Other_EoT, Result):
      Result_EoT = Result.unification_chain_end()
      # To make the PyCharm type checker happy.
      assert isinstance(Result_EoT, PySequence)
      return Result_EoT

  def __getitem__(self, key: Union[int, slice]):
    return self.args[key] if isinstance(key, int) else self.__class__(self.args[key])

  def __len__(self):
    return len(self.args)

  def __str__(self):
    (left, right) = {list: ('[', ']'), tuple: ('(', ')')}[self.functor]
    values_string = self.values_string(self.args)
    result = f'{left}{values_string}{right}'
    return result

  def get_py_value(self) -> tuple:
    return self.functor(arg.get_py_value() for arg in self.args)

  def has_contiguous_sublist(self, As: List):
    """ Can As be unified with a contiguous segment of this list? """
    (len_As, len_self) = (len(As), len(self))
    if len_As == 0:
      yield  # Succeed
    elif len_As > len_self:
      return  # Fail.
    else:
      for i in range(len_self - len_As + 1):
        # Succeed for each segment of self that can be unified with As.
        # This is the same strategy used in the LinkedList version. Just much more straightforward.
        for _ in unify_sequences(As, self.args[i:i+len_As]):
          yield

  def head(self):
    return self[0]

  def tail(self) -> PySequence:
    return self.__class__(self.args[1:])

  def to_python_list(self) -> list:
    return [*self.args]


class PyList(PySequence):
  def __init__(self, initialElements: list):
    super().__init__( list, initialElements )


class PyTuple(PySequence):
  def __init__(self, initialElements: tuple):
    super( ).__init__( tuple, initialElements )


@euc
def append(Xs: Union[PySequence, Var], Ys: Union[PySequence, Var], Zs: Union[PySequence, Var]):
  """
    append([], Ys, Zs).
    append([X|Xs], Ys, [X|Zs]) :- append(Xs, Ys, Zs).

    See discussion in linked_list version.

    This version assumes we are working with Python lists or tuples, i.e., no uninstantiated tails.
  """

  if isinstance(Zs, Var) and ( isinstance(Xs, Var) or isinstance(Ys, Var) ):
      # Can't have Xs or Ys Var if Zs is Var.
      return

  if isinstance(Zs, Var):
    ListType = type(Xs)
    # Make Zs a list of Var's of length len(Xs) + len(Ys)
    # After this unification, Zs will still be a Var, but
    # by the time we reach this point again, Zs will refer
    # to its trail end, which is not a Var.
    for _ in unify(Zs, ListType(n_Vars(len(Xs) + len(Ys)))):
      yield from append(Xs, Ys, Zs)
    return

  # We now know that: Zs is not a Var -- although it may be a sequence of Vars.
  # Divide up its length among Xs and Ys
  ListType = type(Zs)
  len_Zs = len(Zs)
  for i in range(len_Zs + 1):
    # If Xs or Ys are already instantiated to some fixed length Sequence, unify will fail when given the wrong length.
    for _ in unify_pairs([ (Xs, ListType(n_Vars(i))),
                           (Ys, ListType(n_Vars(len_Zs - i)))  ]):
      # Although the lengths of Xs and Ys vary with i,
      # Xs, Ys, and Zs are all of fixed lengths in which len(Xs) + len(Ys) = len(Zs).
      # Concatenate Xs and Ys and then unify the concatenation with Zs.
      XYs = [*Xs.unification_chain_end().args, *Ys.unification_chain_end().args]
      yield from unify_sequences(XYs, Zs.args)
      # for _ in unify_sequences(XYs, Zs.args):
      #   yield


if __name__ == '__main__':

  print(PyTuple( (1, 2, 3) ))
  print(PyList( [1, 2, 3] ))

  Xs = PyList(['Python'])  # list(map(PyValue, range(3)))
  Ys = PyList([PyValue('Q')])  # [PyValue(i+3) for i in range(3)]
  Zs = Var()

  print(f'\nappend({Xs}, {Ys}, {Zs})')
  for _ in append(Xs, Ys, Zs):
    print(f'\tXs: {Xs}, Ys: {Ys}, Zs: {Zs}')

  Xs = Var( )
  Ys = Var( )
  Zs = PyList(list(range(5)))

  print(f'\nappend({Xs}, {Ys}, {Zs})')
  for _ in append(Xs, Ys, Zs):
    print(f'\tXs: {Xs}, Ys: {Ys}, Zs: {Zs}')

  Xs = Var( )
  Ys = Var( )
  Zs = PyTuple(tuple(range(5)))

  print(f'\nappend({Xs}, {Ys}, {Zs})')
  for _ in append(Xs, Ys, Zs):
    print(f'\tXs: {Xs}, Ys: {Ys}, Zs: {Zs}')

  X = tuple(n_Vars(15))
  Y = X[4:8]
  print(f'\nX: {PyTuple(X)}, Y: {PyTuple(Y)}')
  for _ in unify_sequences(Y, tuple(map(PyValue, ['A', 'B', 'C', 'D']))):
    print(f"unify_sequences(Y, tuple(map(PyValue, ['A', 'B', 'C', 'D']))) => X: {PyTuple(X)}, Y: {PyTuple(Y)}")

  B = tuple(n_Vars(8))
  print(f'\nB: {PyTuple(B)}')

  for _ in unify(B[3], PyValue('XYZ')):
    print(f"unify(B[3], PyValue('XYZ')) => B: {PyTuple(B)}")



# from inspect import getmembers
from typing import Generator

from .logic_variables import euc, PyValue, unify, unify_pairs, Var


class Bool_Yield_Wrapper:
  """
  Objects of this class are created to serve as generators. It takes two steps.
  Suppose gen is a generator, i.e.,

  def gen():
      yield

  Such generators are decorated with @bool_yield_wrapper, which returns a function that will create an
  instance of this class when invoked.

  That function is invoked by calling the original generator, e.g.,

  runnable_gen = bool_yield_wrapper(gen)(...)

  runnable_gen will be an instance of Bool_Yield_Wrapper, this class. It is run through its has_more( ) method. E.g.
    while runnable_gen.has_more( ):
        <do something>

  The original gen function must be such that it returns its value through one of its parameters.
  To use the append function as an example,

  @bool_yield_wrapper
  def append(_Xs: Union[PrList, Var], Ys: Union[PrList, Var], _Zs: Union[PrList, Var]):
      ....

  runnable_append = append(Xs, Ys, Zs) # Presumably one or more of Xs, Ys, and Zs is uninstantiated.

  while runnable_append.has_more( ): ... will instantiate the uninstantiated variables.

    (Xs, Ys, Zs) = (Var(), Var(), PrList( [1, 2, 3] )
    runnable_append = append(Xs, Ys, Zs)
    while runnable_append.has_more( ):
        print(f'Xs = {Xs}\nYs = {Ys}\nZs = {Zs}')
        # Prints the various instantiations of Xs and Ys, which, when concatenated, produce Zs.

  These Bool_Yield_Wrapper objects are used in "with" and "while" statements. The current strategy is to
  have a with-statement and embedded while-statement for each generator. For example:

      with append(Xs, Ys, Zs) as gen:
          while gen.has_more():
              <do something>

  The body of the while-loop will repeat each time the generator succeeds.
  """
  def __init__(self, gen):
    self.done = False
    self.gen = gen
    self.name = gen.__name__

  # __enter__ and __exit__ are required for "with"-statements
  def __enter__(self):
    return self

  def __exit__(self, type, value, traceback):
    """
    The three parameters get values if an Exception occurs inside the "with" block, in which case
    __exit__ is passed those properties of the Exception. If __exit__ returns True, it is assumed the
    Exception has been handled. Otherwise, it is assumed the Exception has not been handled and is
    raised outside the "with" block. If no Exception occurs, the values passed are all None, and
    the __exit__ returned value ignored.
    :param type: of Exception, if any; otherwise None.
    :param value: of Exception, if any; otherwise None.
    :param traceback: of Exception, if any; otherwise None.
    :return: Whether Exception has been handeled or should be re-raised.
    """
    pass

  def __iter__(self):
    return self

  def __next__(self):
    """ Can be called next(runnable_append) as well as runnable_append.has_more() """
    if self.done:
      return False
    else:
      try:
        next(self.gen)
        return True
      except StopIteration:
        self.done = True
        return False

  def has_more(self):
    """
    Called only by a while embedded within a with.
    """
    return self.__next__( )


def bool_yield_wrapper(gen):
  """
  A decorator. Generates the Bool_Yield_Wrapper object. See is_even_3, below
  Can also be used explicitly as a function. See is_even_5, below.
  :param gen: the function (not a call to the function) being decorated.
  """
  def wrapped_func(*args, **kwargs):
    return Bool_Yield_Wrapper(gen(*args, **kwargs))

  return wrapped_func


def bool_to_sf(b: bool) -> Generator[None, None, None]:
  """ 
  Turns a boolean condition into a Generator, which succeeds/fails
  if the boolean condition is True/False
  """
  if b:
    yield


def fails(f):
  """
  Applied to a function so that the resulting function succeeds iff the original fails.
  Note that it is applied to the function itself, not to a function call.
  Similar to a decorator but applied explicitly when used.
  """
  def fails_wrapper(*args, **kwargs):
    for _ in f(*args, **kwargs):
      return  # Fail if f succeeds
    yield     # Succeed if f fails.

  return fails_wrapper


def forall(gens):
  """
  Succeeds if all generators in the gens list succeed. The elements in the gens list
  are embedded in lambda functions to avoid premature evaluation.
  """
  if not gens:
    # They have all succeeded.
    yield
  else:
    # Get gens[0] and evaluate the lambda expression to get a fresh iterator.
    # If it succeeds, run the rest of the generators in the list.
    for _ in gens[0]( ):
      yield from forall(gens[1:])


def forany(gens):
  """
  Succeeds if any of the generators in the gens list succeed. On "back-up," tries them all.
  """
  for gen in gens:
    # Get the gens; evaluate them to extract them from lambda; and run them.
    # Succeed if any of them succeed.
    # Can probably not do the lambda trick here, but keeping it makes it
    # parallel to forall. Succeed if any of them succeed.
    # Try them all even if earlier ones succeed. 
    # for _ in gen( ):
    #   yield
    yield from gen( )


def trace(x, succeed=True, show_trace=True):
  """
  Can be included in a list of generators (as in forall and forany) to see where we are.
  The second argument determines whether it succeeds or fails.
  When included in a list of forall generators, succeed should be set to True so that
  it doesn't provent forall from succeeding.
  When included in a list of forany generators, succeed should be set to False so that forany
  will just go on the the next generator one and won't take this one as an extraneous successes.
  """
  if show_trace:
    print(x)
  if succeed:
    yield


def would_succeed(f):
  """
  Applied to a function so that the resulting function succeeds/fails iff the original succeeds/fails.
  If the original succeeds, this also succeeds but without binding any variables.
  Similar to a decorator but applied explicitly when used.
  """
  def would_succeed_wrapper(*args, **kwargs):
    succeeded = False
    for _ in f(*args, **kwargs):
      succeeded = True
      # Do not yield in the context of f succeeding.
      # So un-unify any unifications that occurred in f.

    if succeeded:
      yield  # Succeed if f succeeded.
    # else:
    #   pass   # Fail if f failed

  return would_succeed_wrapper


def yield_to_bool(f):

  def y_to_b_wrapper(*args, **kwargs):
    for _ in f(*args, **kwargs):
      return True
    return False

  return y_to_b_wrapper


if __name__ == '__main__':

  # The following examples illustrate various ways to call a generator
  # and to pass results back through logic variable arguments.

  # Yields when i is even.
  @euc
  def is_even_1(i: int) -> Generator[None, None, None]:
    for _ in unify(PyValue(True), PyValue(i % 2 == 0)):
        yield

  # If is_even_1(i) fails to unify, for _ in is_even_1(i) fails.
  # It works similarly to an if condition.
  evens_1 = [i for i in range(5) for _ in is_even_1(i)]
  print(f'\n1. evens_1: {evens_1}')    # => 1. evens_1: [0, 2, 4]

  # Same as is_even_1 but includes the range generator.
  def is_even_2(n: int, Res: Var) -> Generator[None, None, None]:
    for i in range(n):
        for _ in unify_pairs([ (PyValue(i % 2 == 0), PyValue(True)),
                               (PyValue(i), Res)
                               ]):
          yield

  # Can reuse this variable in all examples.
  # After each example, it is left uninstantiated.
  Result = Var()

  # This version uses the function directly as a generator.
  evens_2 = [Result.get_py_value() for _ in is_even_2(7, Result)]
  print(f'2. evens_2: {evens_2}')    # => 2. evens_2: [0, 2, 4, 6]

  evens_3 = []
  # Here we apply the decorator function explicitly.
  # bool_yield_wrapper makes it possible to use the generator with .has_more() as a boolean.
  # The has_more() function, defined in the BoolYIeldWrapper class, both calls next() on
  # the generator and returns True or False depending on whether next() has succeeded. If
  # next() succeeds, the next value is unified with Result. The next value is also stored
  # within the BoolYIeldWrapper object as self.next, and can be retrieved that way.
  with bool_yield_wrapper(is_even_2)(9, Result) as is_even_gen_3:
    while is_even_gen_3.has_more( ):
      evens_3.append(Result.get_py_value())
  print(f'3. evens_3: {evens_3}')    # => 3. evens_4: [0, 2, 4, 6, 8]

  @bool_yield_wrapper
  def is_even_2_decorated(n: int, Res: Var) -> Bool_Yield_Wrapper:
    for i in range(n):
        for _ in unify_pairs( [(PyValue(i % 2 == 0), PyValue(True)),
                              (PyValue(i), Res)
                               ]):
          yield

  evens_4 = []
  # Create the generator in a separate step.
  is_even_gen_4 = is_even_2_decorated(11, Result)
  while is_even_gen_4.has_more( ):
    evens_4.append(Result.get_py_value())
  print(f'4. evens_4: {evens_4}')    # => 4. evens_4: [0, 2, 4, 6, 8, 10]

  evens_5 = []
  # Create the generator in a 'with' statement.
  with is_even_2_decorated(13, Result) as is_even_gen_5:
    while is_even_gen_5.has_more( ):
      evens_5.append(Result.get_py_value())
  print(f'5. evens_5: {evens_5}')    # => 5. evens_5: [0, 2, 4, 6, 8, 10, 12]

  print('\nEnd of test')



from __future__ import annotations
from functools import wraps
from inspect import isgeneratorfunction
from numbers import Number
from typing import Any, Iterable, List, Optional, Sequence, Sized, Tuple, Union

"""
Developed by Ian Piumarta as the "unify" library (http://www.ritsumei.ac.jp/~piumarta/pl/src/unify.py) for a
course on programming languages (http://www.ritsumei.ac.jp/~piumarta/pl/index.html). 

See discussion on the GitHub page: https://github.com/RussAbbott/pylog
"""

"""
The pylog core, this file contains the logic variable and data structure classes and the unify function.
"""


def euc(f):
  """
  A decorator that takes unification_chain_end() of all Var arguments.
  comment...
  """

  def var_unification_chain_end(v):
    return v.unification_chain_end() if isinstance(v, Var) else v

  def arg_Vars_unification_chain_ends(args):
    args_unification_chain_ends = (var_unification_chain_end(arg) for arg in args)
    return args_unification_chain_ends

  def dict_Vars_unification_chain_ends(dic):
    dic_unification_chain_ends = {k: var_unification_chain_end(v) for (k, v) in dic.items()}
    return dic_unification_chain_ends

  @wraps(f)
  def euc_wrapper_gen(*args, **kwargs):
    args_unification_chain_ends = arg_Vars_unification_chain_ends(args)
    kwargs_unification_chain_ends = dict_Vars_unification_chain_ends(kwargs)
    yield from f(*args_unification_chain_ends, **kwargs_unification_chain_ends)

  @wraps(f)
  def euc_wrapper_non_gen(*args, **kwargs):
    args = arg_Vars_unification_chain_ends(args)
    kwargs_unification_chain_ends = dict_Vars_unification_chain_ends(kwargs)
    return f(*args, **kwargs_unification_chain_ends)

  return euc_wrapper_gen if isgeneratorfunction(f) else euc_wrapper_non_gen


class Term:
  """

                                                     Term  (An abstract logic variable superclass)
                                                       |
                               ------------------------------------------------
                               |                       |                      |
                            PyValue                   Var                 Structure
                               |                                              |
                   --------------------------                            SuperSequence
                    int, float, string, etc.                                  |
                   Any immutable Python value                        -------------------
                                                                     |                 |
                                                                 LinkedList       PySequence
                                                                                       |
                                                                                 -------------
                                                                                 |           |
                                                                               PyList      PyTuple
  """

  term_count = 0

  def __init__(self):
    Term.term_count += 1
    self.term_id = self.term_count

  # @euc Can't use decorators on dunder methods without doing this:
  # (https://stackoverflow.com/questions/55550300/python-how-to-decorate-a-special-dunder-method).
  # Not sure I understand it. Not sure it's worth the trouble.
  def __eq__(self, other: Term) -> bool:
    """
    self == other if either (a) they have the same py_value or (b) are the same variable.
    """
    (self_euc, other_euc) = (self.unification_chain_end(), other.unification_chain_end())
    return self_euc is other_euc or \
           (self is not self_euc or other is not other_euc) and self_euc == other_euc

  def __lt__(self, other: Term) -> bool:
    return str(self) < str(other)

  def __ne__(self, other: Term) -> bool:
    return not (self == other)

  def __str__(self) -> str:
    """
    The str( ) of a Var is (a) the str of its py_value if is_instantiated( ) or (b) its term_id otherwise.
    """
    self_euc = self.unification_chain_end( )
    return f'{self_euc}' if self_euc.is_instantiated( ) or isinstance(self_euc, Structure) else \
           f'_{self_euc.term_id}'

  def get_py_value(self) -> Any:
    return None

  @euc
  def is_instantiated(self) -> bool:
    """ Should never get here since no unification_chain_end is a Term. """
    return False

  def unification_chain_end(self) -> Term:
    return self


def is_immutable(x):
  return (isinstance(x, (Number, str, bool, type(None))) or
          isinstance(x, (frozenset, tuple)) and all(is_immutable(c) for c in x))


class PyValue(Term):

  """ A wrapper class for integers, strings, and other immutable Python value. """

  def __init__(self, py_value: Optional[str, Number] = None ):
    assert is_immutable(py_value), f"Only immutable values are allowed as PyValues. {py_value} is mutable."
    self._py_value = py_value
    super( ).__init__( )

  def __eq__(self, other: Term) -> bool:
    other_euc = other.unification_chain_end()
    return (isinstance(other_euc, PyValue) and
            self.get_py_value() == other_euc.get_py_value() and
            # Don't need to test both.
            self.is_instantiated() and other.is_instantiated)

  def __lt__(self, other: Term) -> bool:
    other_euc = other.unification_chain_end()
    return isinstance(other_euc, PyValue) and self.get_py_value() < other_euc.get_py_value()

  def __str__(self) -> str:
    return f'{self._py_value}'

  # This instantiates a PyValue, which had been None. This is dangerous since it mutates this object.
  def _set_py_value(self, py_value):
    assert is_immutable(py_value), f"Only immutable values are allowed as PyValues. {py_value} is mutable."
    self._py_value = py_value

  def get_py_value(self) -> Any:
    return self._py_value

  @staticmethod
  def get_py_values(Vars: List[Union[Var, PyValue]]):
    return [v.get_py_value( ) for v in Vars]

  def is_instantiated(self) -> bool:
    return self.get_py_value() is not None


class Structure(Term):
  """
  self.functor is the functor
  self.args is a tuple of args
  """
  def __init__(self, term: Tuple = ( None, () ) ):
    self.functor = term[0]
    self.args = tuple(map(ensure_is_logic_variable, term[1:]))
    super().__init__()

  def __eq__(self, other: Term) -> bool:
    other_euc = other.unification_chain_end()
    return (other_euc is self or
            isinstance(other_euc, Structure) and
            self.functor == other_euc.functor and
            len(self.args) == len(other_euc.args) and
            all([selfArg == other_eucArg for (selfArg, other_eucArg) in zip(self.args, other_euc.args)]))

  def __getitem__(self, key: Union[int, slice]):
    return self.args[key]

  # noinspection PySimplifyBooleanCheck
  def __str__(self):
    args_str = self.values_string(self.args)
    result = f'{self.functor}({args_str})'
    return result

  def get_py_value(self) -> Structure:
    py_value_args = [arg.get_py_value() for arg in self.args]
    return Structure( (self.functor, *py_value_args) )

  def is_instantiated(self) -> bool:
    """ A Structure is instantiated if all its args are. """
    args_are_instantiated = all(arg.is_instantiated() for arg in self.args)
    return args_are_instantiated

  @staticmethod
  def values_string(values: Iterable):
    result = ', '.join(map(str, values))
    return result


class StructureItem(Structure):
  """
  A utility class for building and displaying Structure-based items.
  """

  def __init__(self, args, first_arg_as_str_functor=False):
    self.first_arg_as_str_functor = first_arg_as_str_functor
    functor = type(self).__name__.lower( )
    super().__init__( (functor, *map(make_property, args)) )

  def __str__(self):
    all_args_uninstantiated = all(isinstance(arg.unification_chain_end(), Var) for arg in self.args)
    if all_args_uninstantiated:
      # If all the args are uninstantiated, print a simple underscore.
      return '_'
    elif self.first_arg_as_str_functor:
      # If first_arg_as_str_functor, use the first arg as the functor for printing.
      args_str = ', '.join(map(str, self.args[1:]))
      result = f'{self.args[0]}({args_str})'
      return result
    else:
      # Use default Structure __str__( )
      return super().__str__( )


class Var(Term):
  """
  A logic variable
  """

  def __init__(self):
    # self.unification_chain_next points to the next element on the unification_chain, if any.
    self.unification_chain_next = None
    super().__init__()

  def __getattr__(self, item):
    self_euc = self.unification_chain_end()
    if self is not self_euc:
      return self_euc.__getattribute__(item)

  # Apparently __getattr__ is not called for calls to __getitem__ when __getitem__ is missing
  def __getitem__(self, key: Union[int, slice]):
    self_euc = self.unification_chain_end()
    if self is not self_euc and hasattr(self_euc, '__getitem__'):
      return self_euc.__getitem__(key)

  def __len__(self):
    self_euc = self.unification_chain_end()
    # To make PyCharm's type checker happy.
    assert isinstance(self_euc, Sized)
    return None if not hasattr(self_euc, '__len__') or self == self_euc else len(self_euc)

  def _has_unification_chain_next(self) -> bool:
    # Is this the end of the unification_chain?
    return self.unification_chain_next is not None

  @euc
  def get_py_value(self) -> Optional[Any]:
    return self.get_py_value( ) if self.is_instantiated( ) else None

  # Can't use @euc. Generates an infinite recursive loop.
  def is_instantiated(self) -> bool:
    """ A Var is_instantiated if its unification_chain end is_instantiated """
    Trail_End_Var = self.unification_chain_end( )
    return not isinstance(Trail_End_Var, Var) and Trail_End_Var.is_instantiated()

  def unification_chain_end(self):
    """
    return: the Term, whatever it is, at the end of this Var's unification unification_chain.
    """
    return self.unification_chain_next.unification_chain_end( ) if self._has_unification_chain_next( ) else self


# @staticmethod
def ensure_is_logic_variable(x: Any) -> Term:
  """
    Applied to each argument in a Structure.
    Applies PyValue to those that are not already Terms.
    If x is not a logic variable, i.e., an instance of Term, it must be a Python value.
    Wrap it in PyValue. (It must be immutable.)
  """
  return x if isinstance(x, Term) else PyValue(x)


# @staticmethod
def make_property(prop):
  """
    Use in StructureItem -- for puzzles.
    If a property is None, create a Var for it.
    Otherwise apply ensure_is_logic_variable.
  """
  return Var( ) if prop is None else ensure_is_logic_variable(prop)


def n_Vars(n: int) -> List[Var]:
  """ Generate a list of uninstantiated variables of length n. """
  return [Var( ) for _ in range(n)]


# noinspection PyProtectedMember
@euc
def unify(Left: Any, Right: Any):
  """
  Unify two logic Terms.

  The strategy is to keep track of the "unification unification_chain" for all variables.

  The unification unification_chain is a linked list of logic variables, which are all unified.

  The final element on the unification_chain is either
  o a non-Var, in which case the value of all preceding variables is the value of that non-Var, or
  o a Var (which is not linked to any further element), in which case, all variables on the unification_chain
    are unified but do not (yet) have a value.
  """

  # Make sure both Left and Right are logic variables. This allows us to call, e.g, unify(X, 'abc').
  # ensure_is_logic_variable will wrap 'abc' in a PyValue.
  (Left, Right) = map(ensure_is_logic_variable, (Left, Right))

  # If the unification_chain_ends are equal, either because they have the same py_value or Structure or
  # because they are the same (unbound) Var, do nothing. They are already unified.
  # yield to indicate unification success. If Left and Right are both
  # uninstantiated PyValues, Left != Right. (See PyValue.__eq__.)
  if Left == Right:
    yield

  # The rest consists of special cases: both PyValues, both Structures, at least one Var.

  # Case 1. Both are PyValues, and exactly one is instantiated.
  # "Assign" it's value to the other. This is similar to (but simpler than)
  # how we handle two Var's. But instead of building a unification_chain, we "assign"
  # one value to the other.
  elif isinstance(Left, PyValue) and isinstance(Right, PyValue) and \
       (not Left.is_instantiated( ) or not Right.is_instantiated( )) and \
       (Left.is_instantiated( ) or Right.is_instantiated( )):
    (assignedTo, assignedFrom) = (Left, Right) if Right.is_instantiated( ) else (Right, Left)
    assignedTo._set_py_value(assignedFrom.get_py_value( ))
    yield
    # See discussion in unify below for why we do this.
    assignedTo._set_py_value(None)
    # :
    # # If they are both PyValues, treat specially.
    # yield from unify_PyValues(Left, Right)
    
  # Case 2. Both  Structures. They can be unified if
  # (a) they have the same functor and
  # (b) their arguments can be unified.
  elif isinstance(Left, Structure) and isinstance(Right, Structure) and Left.functor == Right.functor:
    yield from unify_sequences(Left.args, Right.args)

  # Case 3. At least one is a Var. Since we use @euc, it's the end of its unification_chain.
  # Make the other an extension of its unification_chain.
  # (If both are Vars, it makes no functional difference which extends which.)
  elif isinstance(Left, Var) or isinstance(Right, Var):
    (pointsFrom, pointsTo) = (Left, Right) if isinstance(Left, Var) else (Right, Left)
    pointsFrom.unification_chain_next = pointsTo
    yield
    # All yields create a context in which more of the program is executed--like
    # the body of a while-loop or a for-loop. A "next()" request asks for alternatives.
    # But there is only one functional way to do unification. So on "backup," unlink the
    # two and exit without a further yield, i.e., fail.

    # This is fundamental! It's what makes it possible for a Var to become un-unified outside
    # the context in which it was unified, e.g., unifying a Var with (successive) members
    # of a list. The first successful unification must be undone before the second can occur.
    pointsFrom.unification_chain_next = None

    
def unify_pairs(tuples: List[Tuple[Any, Any]]):
  """ Apply unify to pairs of terms. """
  # If no more tuples, we are done.
  if not tuples:
    yield
  else:
    # Get the first tuple from the tuples list.
    [(Left, Right), *restOfTuples] = tuples
    # If they unify, go on to the rest of the tuples list.
    for _ in unify(Left, Right):
      yield from unify_pairs(restOfTuples)
    # The preceding is equivalent to the following.
    # for _ in forall([lambda: unify(Left, Right),
    #                  lambda: unify_pairs(restOfTuples)]):
    #   yield


def unify_sequences(seq_1: Sequence, seq_2: Sequence):
  """ Unify simple sequences, e.g., lists or tuples, of Terms. """
  # The two sequences must be the same length.
  if len(seq_1) != len(seq_2):
    return

  # If they are both empty, we are done.
  if len(seq_1) == 0:
    yield

  else:
    # Unify the two first elements. If successful go on to the rest.
    # Note that slice notation is supported.
    for _ in unify(seq_1[0], seq_2[0]):
      yield from unify_sequences(seq_1[1:], seq_2[1:])


if __name__ == '__main__':

  A = 'abc'
  B = Var( )
  C = Var( )
  D = 'def'
  print(f'\nA: {A}; B: {B}; C: {C}; D: {D}')
  print(f'Attempting: unify_pairs([(A, B), (B, C), (C, D)]).  A, B, C, D will all be the same if it succeeds.')
  for _ in unify_pairs([(A, B), (B, C), (C, D)]):
    print(f'b. A: {A}; B: {B}; C: {C}; D: {D}')
  print('As expected, unify_pairs fails -- because A and D have distinct PyValue values.')

  PV1 = PyValue()
  PV2 = PyValue()
  print(f'\nTrying unify({PV1}, {PV2}). '
        f"Should fail because we explicitly don't allow two uninstantiated PV's to unify.")
  for _ in unify(PV1, PV2):
    print("Shouldn't have succeeded.")
  print("Failed, as expected, if nothing before this.")
  print(f'Trying unify({PV1}, 1). Should succeed.')
  for _ in unify(PV1, 1):
    print(f"PV1.is_instantiated(): {PV1.is_instantiated()}: {PV1}")
  print(f"PV1.is_instantiated(): {PV1.is_instantiated()}: {PV1}")

  A = Var( )
  B = Var( )
  C = Var( )
  D = 'def'

  print(f'\n1. A: {A}; B: {B}; C: {C}; D: {D}')
  for _ in unify(A, B):
    print(f'After unify(A, B).  A: {A}; B: {B}; C: {C}; D: {D}')
    for _ in unify(A, C):
      print(f'After unify(A, C). A: {A}; B: {B}; C: {C}; D: {D}')
      for _ in unify(A, D):
        print(f'After unify(A, D). A: {A}; B: {B}; C: {C}; D: {D}')
  print(f'Outside the scope of all unifies: ')
  print(f'            A: {A}; B: {B}; C: {C}; D: {D}')

  print('End first test\n')

  """
  A: abc; B: _2; C: _3; D: def
  Attempting: unify_pairs([(A, B), (B, C), (C, D)]).  A, B, C, D will all be the same if it succeeds.
  As expected, unify_pairs fails.
  
  1. A: _5; B: _6; C: _7; D: def
  2a. After unify(A, B).  A: _6; B: _6; C: _7; D: def
  2b. After unify(A, C). A: _7; B: _7; C: _7; D: def
  2c. After unify(A, D). A: def; B: def; C: def; D: def
  3. Outside the scope of all unifies. A: _5; B: _6; C: _7; D: def
  End first test
  """

  A = Var( )
  B = Var( )
  C = Var( )
  D = 'xyz'

  print(f'1. A: {A}, B: {B}, C: {C}, D: {D}')
  for _ in unify_pairs([(A, B), (B, C)]):
    print(f'2. After unify_pairs([(A, B), (B, C)]):. A: {A}, B: {B}, C: {C}, D: {D}')

    for _ in unify(D, B):
      print('3. After unify(D, B): A: {A}, B: {B}, C: {C}, D: {D}'  # => A.euc: xyz, B.euc: xyz, C.euc: xyz, D.euc: xyz
            )

    print(f'\n4. No longer unified with D. A: {A}, B: {B}, C: {C}')  # => A: xyz, B: xyz, C: xyz, D: xyz
  print(f'5. No longer unified with each other. A: {A}, B: {B}, C: {C}')  # => A: xyz, B: xyz, C: xyz, D: xyz
  print('\nEnd second test\n')

  """
  Expected output

  1. A: _13, B: _14, C: _15
  2. A: _15, B: _15, C: _15
  3. A.euc: _15, B.euc: _15, C.euc: _15
  4. A.euc: xyz, B.euc: xyz, C.euc: xyz, D.euc: xyz

  5. A: _15, B: _15, C: _15
  6. A: _13, B: _14, C: _15

  End second test
  """

  X = Var( )
  Y = Var( )
  Z = Var( )
  print(f'X: {X}, Y: {Y}, Z: {Z}')
  for _ in unify('abc', X):
    print(f'After unify("abc", X): X: {X}, Y: {Y}, Z: {Z}')  # => abc
    for _ in unify(X, Y):
      print(f'After unify(X, Y): X: {X}, Y: {Y}, Z: {Z}')  # => abc
      for _ in unify(Z, Y):
        print(f'After unify(Z, Y): X: {X}, Y: {Y}, Z: {Z}')  # => abc
      print(f'Outside unify(Z, Y): X: {X}, Y: {Y}, Z: {Z}')  # => abc
    print(f'Outside unify(X, Y): X: {X}, Y: {Y}, Z: {Z}')  # => abc
  print('\nEnd third test\n')

  V1 = Var()
  T1 = Structure( ('t', 1, V1, V1))
  V2 = Var()
  V3 = Var()
  T2 = Structure( ('t', V2, V2, V3))

  print(f'V1: {V1}, V2: {V2}, V3: {V3}, ')
  print(f'T1: t(1, V1, V1), T2: t(V2, V2, V3)')
  for _ in unify(T1, T2):
    print('After unify(T1, T2):')
    print(f'V1: {V1}, V2: {V2}, V3: {V3}, ')
    print(f'T1: {T1}, T2: {T2}')
    print('End of fourth test.')

  """
  Expected output
  
  T1: t(1, 1, 1), T2: t(1, 1, 1)
  V1: 1, V2: 1, V3: 1, 
  End of fourth test.

  """

  V4 = Var()
  T4 = Structure( ('t', 1, V4))
  print(f'\nV4: {V4}')
  print(f'T4: t(1, V4)')
  for _ in unify(T4, V4):
    print('After unify(T4, V4):')
    print(f'V4[0]: {V4[0]}')
    print(f'V4[1] is T4: {V4[1] is T4}')
    print(f'V4[1] == T4: {V4[1] == T4}, '
          f'because: V4[1].unification_chain_end() is T4: {V4[1].unification_chain_end() is T4}')
    print('An attempt to print T4 or V4 will produce "RecursionError: maximum recursion depth exceeded"')
    print('\nEnd of fifth test.')

  """
  Expected output

  V4: _23
  T4: t(1, V4)
  After unify(T4, V4):
  V4[0]: 1
  V4[1] is T4: False
  V4[1] == T4: True, because: V4[1].unification_chain_end() is T4: True
  An attempt to print T4 or V4 will produce "RecursionError: maximum recursion depth exceeded"
  
  End of fifth test.

  """

  T5 = Structure( ('g', 1, 2, 3) )
  print(f'\nT5 = Structure( ("g", 1, 2, 3) ): {T5}')
  T6 = Structure( ('t', *range(4), T5, *range(5, 9)) )
  print(f'T6 = Structure( ("t", *range(4), T5, *range(5, 9) ): {T6}')
  print(f'(", ".join(map(str, T6[3:8]))): ({", ".join(map(str, T6[3:8]))})')
  print(f'tuple(x.get_py_value( ) for x in T6[4][1:3]): { tuple(x.get_py_value() for x in T6[4][1:3]) }')
  print('\nEnd of sixth test.')

  """
  Expected output
  
  T5 = Structure( ("g", 1, 2, 3) ): g(1, 2, 3)
  T6 = Structure( ("t", *range(4), T5, *range(5, 9) ): t(0, 1, 2, 3, g(1, 2, 3), 5, 6, 7, 8)
  (", ".join(map(str, T6[3:8]))): (3, g(1, 2, 3), 5, 6, 7)
  tuple(x.get_py_value( ) for x in T6[4][1:3]): (2, 3)
  
  End of sixth test.
  """



from __future__ import annotations
from typing import List, Union

from ..control_structures import forany, forall
from ..logic_variables import euc, Structure, unify, Term, Var


class SuperSequence(Structure):
  """
  A subclass of Structure that can serve as a superclass for both PySequence and LinkedList
  Declares a number of abstract methods.
  """

  def __getitem__(self, key: Union[int, slice]) -> Union[SuperSequence, Term]:
      pass

  def __len__(self):
    pass

  def has_contiguous_sublist(self, As):
    pass

  def head(self) -> Term:
    pass

  def is_empty(self) -> bool:
    # An empty list has no args, i.e., no head or tail.
    # A linked list with a variable tail will also have a head, which itself may be a Var.
    # There is no way to represent a LinkedList is nothing but a variable tail. It is just a simple Var.
    return not self.args

  def to_python_list(self) -> list:
    return []

  def tail(self) -> SuperSequence:
    pass


@euc
def is_contiguous_in(As: List, Zs: SuperSequence):
  """ Can As be unified with a segment of Zs? """
  yield from Zs.has_contiguous_sublist(As)


@euc
def is_a_subsequence_of(As: List, Zs: SuperSequence):
  """
  As may be spread out in Zs but must be in the same order as in Zs.
  """
  if not As:
    # If no more As to match, we're done. Succeed.
    yield

  elif not Zs:
    # If no more Zs to match the remaining As, fail.
    return

  else:
    for _ in forany([
                     # Match As[0] and Zs[0]; go on to is_a_subsequence_of(As[1:], Zs[1:])
                     lambda: forall([lambda: unify(As[0], Zs[0]),
                                     lambda: is_a_subsequence_of(As[1:], Zs[1:])]),
                     # Whether or not we matched As[0] and Zs[0] above, try is_a_subsequence_of(As, Zs[1:])
                     lambda: is_a_subsequence_of(As, Zs[1:])
                     ]):
      yield


@euc
def member(E: Term, A_List: Union[List, SuperSequence, Var]):
  """
  Is E in A_List?
  """
  # if isinstance(A_List, list):
  #   yield from member_python_list(E, A_List)
  #
  # # If A_List is empty, it can't have a member. So fail.
  if A_List.is_empty():
    return

  # The following is an implicit 'or'. Either unify E with A_List.head() or call member(E, A_List.tail()).

  # The first case is easy.
  # for _ in unify(E, A_List.head( )):
  #   yield
  yield from unify(E, A_List.head( ))

  # The second case--member(E, A_List.tail())--is trickier.
  # Since A_List may be an open-ended LinkedList, A_List.tail() may be a Var.
  # In that case, we must first instantiate A_List.tail() to LinkedList( (Var, Var) ).
  A_List_Tail = A_List.tail()
  # Create A_List_New_Tail to be unified with A_List_Tail.
  # A_List_New_Tail will be A_List_Tail in most cases.
  # But if isinstance(A_List_Tail, Var), make A_List_New_Tail = LinkedList( (Var( ), Var( )) ).
  # In either case, unify A_List_Tail with A_List_New_Tail and call member(E, A_List_New_Tail).
  # An issue is that we can't import LinkedList since that would create an import cycle.
  # Instead use type(A_List), which will be LinkedList if A_List_Tail is a Var.
  A_List_New_Tail = type(A_List)((Var( ), Var( ))) if isinstance(A_List_Tail, Var) else A_List_Tail
  # If A_List_New_Tail is A_List_Tail, this unify does nothing.
  for _ in unify(A_List_New_Tail, A_List_Tail):
    yield from member(E, A_List_New_Tail)


@euc
def members(Es: List, A_List: SuperSequence):
  """ Do all elements of es appear in A_List (in any order). """
  # No more to look up. We're done.
  if not Es:
    yield
  else:
    for _ in member(Es[0], A_List):
      yield from members(Es[1:], A_List)


def next_to(E1: Term, E2: Term, Es: SuperSequence):
  """ Are E1 and E2 are next to each other in Es. """
  for _ in forany([
                   lambda: is_contiguous_in([E1, E2], Es),
                   lambda: is_contiguous_in([E2, E1], Es),
                   ]):
    yield


def reversed(A_List: SuperSequence) -> SuperSequence:
  A_List_to_python_list = A_List.to_python_list()
  reversed_A_List = A_List_to_python_list[::-1]
  return type(A_List)(reversed_A_List)


from __future__ import annotations
from typing import Any, List, Optional, Tuple, Union

from ..logic_variables import ensure_is_logic_variable, euc, PyValue, n_Vars, Term, unify, unify_pairs, Var
from ..sequence_options.super_sequence import is_a_subsequence_of,  member, SuperSequence


class LinkedList(SuperSequence):
  """
  A special Structure for Prolog (Linked) Lists.
  self.functor = 'linkedList'
  self.args is in general a tuple of args. In this case,
  self.args[0] = head(list)
  self.args[1] = LinkedList(tail(list))

  When used as a constructor, the argument must be either a Python list or a (Head, Tail) tuple.
  In the second case, Tail must be a LinkedList or a Var.
  """
  def __init__(self, list_or_tuple: Union[list, str, Term, tuple], tail: Optional[Term] = None ):
    # args will either have two elements or none -- if we are creating an empty list.
    if tail is None:
      args = self.args_from_pyList(list_or_tuple) if isinstance(list_or_tuple, list) else \
             list_or_tuple if isinstance(list_or_tuple, tuple) else \
             self.args_from_pyList(list(list_or_tuple))  # Run list_or_tuple to get a list if it's a generator.
    else:
      args = (list_or_tuple, tail)
    super().__init__( ('linkedList', *args) )

  def __getitem__(self, key: Union[int, slice]):
    (prefix, tail) = self.prefix_and_tail( )
    stop = key if isinstance(key, int) else \
           key.stop if isinstance(key.stop, int) else \
           len(self) if key.stop is None else \
           None
    if len(prefix) >= stop:
      slice_elements = prefix[key]
      if isinstance(key, int):
        return slice_elements
      else:
        return LinkedList(slice_elements)
    if not isinstance(tail, Var):
      return None
    template = LinkedList(n_Vars(stop))
    for _ in unify(self, template):
      (prefix, _) = template.prefix_and_tail()
      return LinkedList(prefix[key])

  def __len__(self):
    return len(self.to_python_list())

  def __str__(self):
    (prefix, tail) = self.prefix_and_tail( )
    valuesString = self.values_string(prefix)
    result = f'[{valuesString}' + (f' | {tail}]' if isinstance(tail, Var) else ']' )
    return result

  @staticmethod
  # pyList is a standard Python list.
  def args_from_pyList(pyList: list):
    """
    Converts a standard Python list into the appropriate args, i.e., (head, tail), for a LinkedList
    :param pyList: a standard Python list
    :return: (nested) args for a LinkedList, i.e, (head, tail) of pylist.
    """
    # return ( ) if not pyList else ( Term.ensure_is_logic_variable(pyList[0]), LinkedList(pyList[1:]) )
    return ( ) if not pyList else ( ensure_is_logic_variable(pyList[0]), LinkedList(pyList[1:]) )

  def get_py_value(self):
    args_list = self.to_python_list()
    py_value_args = [arg.get_py_value() for arg in args_list]
    return py_value_args

  def has_contiguous_sublist(self, As: List):
    """ Can As be unified with a segment of this list? """
    # Initially, As is a standard Python list. Make it a LinkedList.
    As = LinkedList(As)
    (len_As, len_self) = (len(As), len(self))
    if len_As == 0:
      yield  # Succeed once.
    elif len_As > len_self:
      return  # Fail.
    else:
      (Xs, Ys) = n_Vars(2)
      # Succeed if we can find a way to divide self into Xs and Ys so that As is an initial sublist of Ys.
      # for _ in forall([lambda: append(Xs, Ys, self),
      #                  lambda: append(As, Var( ), Ys)
      #                  ]):
      #   yield
      for _ in append(Xs, Ys, self):
        yield from append(As, Var( ), Ys)

  def head(self) -> Term:
    return self.args[0]

  def prefix_and_tail(self) -> Tuple[List[Term], Any]:
    """ Get the initial list of objects and either the tail if it is a Var or [] if it is not a Var. """
    if self.is_empty():
      return ([], [])
    else:
      Tail_EoT = self.tail().unification_chain_end()
      if isinstance(Tail_EoT, LinkedList):
        (tail_prefix, Tail_Tail) = Tail_EoT.prefix_and_tail( )
        return ([self.head()] + tail_prefix, Tail_Tail)
      else:
        return ([self.head()], Tail_EoT)

  def tail(self) -> Union[LinkedList, Var]:
    return self.args[1]

  def to_python_list(self) -> Union[list, Tuple[list, Var]]:
    (prefix, tail) = self.prefix_and_tail()
    return (prefix, tail) if isinstance(tail, Var) else prefix


emptyLinkedList = LinkedList([])


@euc
def append(Xs: Union[LinkedList, Var], Ys: Union[LinkedList, Var], Zs: Union[LinkedList, Var]):
  """
    append([], Ys, Ys).
    append([X|Xs], Ys, [X|Zs]) :- append(Xs, Ys, Zs).

  Note that this could just have well have been written:
    append([Z|Xs], Ys, [Z|Zs]) :- append(Xs, Ys, Zs).
  or
    append([W|Xs], Ys, [W|Zs]) :- append(Xs, Ys, Zs).

  We don't have to implicitly favor the Xs.

  No matter what the variable is called, this is really unifying Xs_head and Zs_head.

  append/3 converts between: Xs + Ys <--> Zs

  It does this by moving elements between the Xs and the Zs.
  If we are moving from Zs to Xs + Ys, there will be multiple answers depending on
  how much of Zs is shifted before we let Ys be the rest.

  The function consists of two "clauses" (see prolog definition above), which are tried in sequence.
  
  Clause 1
  This is the step in which we set Xs to [] and set Ys to Zs. No recursive call.
  At this point, earlier recursion calls will have shifted elements between Xs and Zs.
  But those are not visible at this level.

  Clause 2. The recursive step.
  Move an element between Xs and Zs and call append recursively with Xs_Tail and Zs_Tail.
  Since an empty list cannot unify with a LinkedList that has a head and a tail, if either
  Xs or Zs is empty, unify_pairs will fail.

  The actual code is quite short. It's very similar to the prolog code.
  """

  # Corresponds to append([], Ys, Ys).
  yield from unify_pairs([(Xs, emptyLinkedList), (Ys, Zs)])

  # Corresponds to append([X | Xs], Ys, [X | Zs]): - append(Xs, Ys, Zs).
  (XZ_Head, Xs_Tail, Zs_Tail) = n_Vars(3)
  for _ in unify_pairs([(Xs, LinkedList(XZ_Head, Xs_Tail)),
                        (Zs, LinkedList(XZ_Head, Zs_Tail))]):
    yield from append(Xs_Tail, Ys, Zs_Tail)


if __name__ == '__main__':

  print( LinkedList([]) == LinkedList(()))

  print(emptyLinkedList)
  E = PyValue(3)
  for _ in member(E, emptyLinkedList):
    print(f'Error: should not get here.')

  A = LinkedList(Var( ), Var( ))
  A112 = A[4:11:2]
  A37 = A[3:7]
  print(f'\nA: {A}\nA[3:7]: {A37}\nA[4:11:2]: {A112}')
  for _ in unify(A37, LinkedList('ABCD')):
    print(f'\nA: {A}\nA[3:7]: {A37}\nA[4:11:2]: {A112}')
    print( )

  print(f'A[:4]: {A[:4]}')
  A_tail = A.tail()
  print(f'A.tail()[:3]: {A_tail[:3]}')
  print(f'A.tail().tail()[:2]: {A.tail().tail()[:2]}')

  print(f'\nemptyLinkedList: {emptyLinkedList}')

  Xs = LinkedList([*range(10)])
  print(f'\nSlices: Xs: {Xs}')
  print(f'\tXs[2:-1:2]: {Xs[2:-1:2]}')
  print(f'\tXs[-1:2:-2]: {Xs[-1:2:-2]}')
  print(f'\tXs[5]: {Xs[5]}')
  """
  Slices: Xs: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    tXs[2, -1, 2]: [2, 4, 6, 8]
    Xs[-1, 3, -1]: [9, 7, 5, 3]
    Xs[5]: 5
  """
  Xs = list(map(PyValue, range(9)))
  Sub_Xs = [PyValue(1), Var( ), PyValue(4), Var( ), PyValue(8)]
  print(f'\nLinkedList(Sub_Xs): {LinkedList(Sub_Xs)}, LinkedList(Xs): {LinkedList(Xs)}')
  for _ in is_a_subsequence_of(Sub_Xs, LinkedList(Xs)):
    print(f'\tLinkedList(Sub_Xs): {LinkedList(Sub_Xs)}')
  Sub_Xs = [PyValue(1), Var( ), PyValue(8), Var( ), PyValue(7)]
  print(f'\nLinkedList(Sub_Xs): {LinkedList(Sub_Xs)}, LinkedList(Xs): {LinkedList(Xs)}')
  for _ in is_a_subsequence_of(Sub_Xs, LinkedList(Xs)):
    print(f'\tLinkedList(Sub_Xs): {LinkedList(Sub_Xs)}')

  (Start, Pivot, End) = n_Vars(3)
  Zs = LinkedList( list(range(10)) )

  L = LinkedList( n_Vars(3) )
  print(f'\nGiven: L = LinkedList( n_Vars(3) ): {L}' )
  E = Var( )
  for _ in unify(L, LinkedList([1, 2, 3])):
    print(f'Given: E: {E}, L: {L}')

    print(f'?- LinkedList.member(E, L)')
    for _ in member(E, L):
      print(f'E = {E}')

  A_List = LinkedList( [*map(PyValue, [1, 2, 3, 2, 1])] )
  E_Sub = [PyValue(2), Var( )]
  linkedlistE_Sub = LinkedList(E_Sub)
  print(linkedlistE_Sub)
  print(f'\nE_Sub: {linkedlistE_Sub}\n?- is_contiguous_in(E_Sub, {A_List})')
  for _ in A_List.has_contiguous_sublist(E_Sub):
    print(f'E_Sub: {linkedlistE_Sub}')

  E_Sub = [Var( ), PyValue(2)]
  linkedlistE_Sub = LinkedList(E_Sub)
  print(f'\nE_Sub: {linkedlistE_Sub}\n?- is_contiguous_in(E_Sub, {A_List})')
  for _ in A_List.has_contiguous_sublist(E_Sub):
    print(f'E_Sub: {linkedlistE_Sub}')

  X = PyValue('abc')
  Z1 = Var( )
  Y1 = LinkedList([X, Z1])

  Z2 = Var( )
  Z3 = Var( )
  Y2 = LinkedList([Z2, Z3])

  print(f'\n1. Y1: {Y1}, Y2: {Y2}')
  for _ in unify_pairs([(Y1, Y2), (Z2, Z3)]):
    print(f'2. Y1: {Y1}, Y2: {Y2}')
  print(f'3. Y1: {Y1}, Y2: {Y2}')

  print('\nAgain, switching the order of the unifications')
  print(f'4. Y1: {Y1}, Y2: {Y2}')
  for _ in unify_pairs([(Y1, Y2), (Z2, Z3)]):
    print(f'5. Y1: {Y1}, Y2: {Y2}')
  print(f'6. Y1: {Y1}, Y2: {Y2}')
  print('End test')

  """
  Expected output
  
  emptyLinkedList: []
  LinkedList([1, 2, 3]): [1, 2, 3]
  
  1. Y1: [abc, _13], Y2: [_17, _17]
  2. Y1: [abc, abc], Y2: [abc, abc]
  3. Y1: [abc, _13], Y2: [_17, _17]
  
  End test
  """

  # Append examples
  # Run "forward:" Xs + Ys --> Zs
  # Only one way to do this.
  (Xs, Ys, Zs) = (LinkedList([0, 1]), LinkedList([2, 3, 4]), Var())
  print(f'\nGiven: Zs: {Zs}\n?- append({Xs}, {Ys}, Zs)')
  for _ in append(Xs, Ys, Zs):
    print(f'Zs = {Zs}')

  # Run "backward:" Zs --> Xs + Ys
  # Multiple ways to split up Zs between Xs and Ys
  (Xs, Ys, Zs) = (Var(), Var(), LinkedList([1, 2, 3]))
  print(f'\nGiven: Xs: {Xs}, Ys: {Ys}, Zs: {Zs}\n?-  append(Xs, Ys, {Zs};')
  for _ in append(Xs, Ys, Zs):
    print(f'Xs = {Xs}\nYs = {Ys}\nZs = {Zs}\n')

  (Xs, Ys, Zs) = (LinkedList([1, 2, 3]), Var(), Var())
  print(f'\nGiven: Xs: {Xs}, Ys: {Ys}, Zs: {Zs}')
  print(f'?- append({Xs}, Ys, Zs); ')
  for _ in append(Xs, Ys, Zs):
    print(f'nlXs = {Xs}')
    print(f'Ys = {Ys}')

  """
  Expected output:
  
  > append([0, 1], [2, 3, 4], Zs); (Zs: _14)
  Zs = [0, 1, 2, 3, 4]
  Zs = _14
  """

  (Xs, Ys, Zs) = (Var(), Var(), LinkedList([1, 2, 3, 4]))
  print(f'\nGiven: Ys: {Ys}, Zs: {Zs}\n?- append({Xs}, Ys, Zs); ')
  for _ in append(Xs, Ys, Zs):
    print(f'Ys = {Ys}')
    print(f'Zs = {Zs}')

  """
  
  > append(Xs, Ys, [0, 1, 2, 3, 4]); (Xs: _36, Ys: _37)
  Xs = []
  Ys = [0, 1, 2, 3, 4]
  
  Xs = [0]
  Ys = [1, 2, 3, 4]
  
  Xs = [0, 1]
  Ys = [2, 3, 4]
  
  Xs = [0, 1, 2]
  Ys = [3, 4]
  
  Xs = [0, 1, 2, 3]
  Ys = [4]
  
  Xs = [0, 1, 2, 3, 4]
  Ys = []
  
  
  > append([1, 2, 3], Ys, Zs); (Ys: _122, Zs: _123)
  Ys = _139;
  Zs = ([], 1, ([], 2, ([], 3, _139)))
  
  """

  print(f'\n?- LinkedList([Var()]).is_instantiated(): {LinkedList([Var( )]).is_instantiated( )}')
  A = LinkedList([1, 2, 3])
  B_Head = Var( )
  B_Tail = Var( )
  B = LinkedList(B_Head, B_Tail)

  print(f'1. A: {A}, B: {B}')
  for _ in unify(A, B):
    print(f'2. A: {A}, B: {B}')

    B_Tail_TrailEnd = B_Tail.unification_chain_end( )
    C = LinkedList([0, *B_Tail_TrailEnd.get_py_value( )])
    print(f'3. C: {C}')

    D2 = Var( )
    D3 = Var( )
    D = LinkedList([D3, D2, D3])

    print(f'4a. B: {B}, D: {D}')
    for _ in unify(D.tail(), B.tail()):
      print(f'4b. unify(D.tail(), B.tail() => B: {B}, D: {D}')
    for _ in unify(D.head(), B):
      print(f'4c. unify(D.head(), B) => B: {B}, D: {D}')
    E = Var( )
    print(f'5a. B: {B}, E: {E}')
    for _ in unify(E, B):
      # Since E is a Var, must take its unification_chain_end to get something that has a head and tail.
      E_EoC = E.unification_chain_end( )
      assert isinstance(E_EoC, LinkedList)
      print(f'5b. unify(E, B) => E: {E}, E_EoT.head(): {E_EoC.head( )}, E_EoT.tail(): {E_EoC.tail( )}\n')

  # The empty LinkedList is a LinkedList ith no arguments.
  head = PyValue('head')
  Unclosed_List1 = LinkedList(head, Var( ))
  print(f'6. Unclosed_List1: {Unclosed_List1}, len(Unclosed_List1): {len(Unclosed_List1)}')
  Unclosed_List2 = LinkedList(Var( ), Var( ))
  print(f'7. Unclosed_List2: {Unclosed_List2}, len(Unclosed_List2): {len(Unclosed_List2)}')
  for _ in unify(LinkedList([*range(5)]), Unclosed_List2):
    print(f'8. unify(LinkedList([*range(5)]), Unclosed_List2) => ' 
          f'Unclosed_List2: {Unclosed_List2}; len(Unclosed_List2): {len(Unclosed_List2)}')

  Unclosed_List3 = LinkedList(Var( ), Var( ))
  print('\nStarting to call member on an open-ended list.')
  limit1 = 4
  for _ in member(PyValue(5), Unclosed_List3):
    if limit1 <= 0:
      break
    limit1 -= 1
    print(Unclosed_List3)
    limit2 = 2
    for _ in member(PyValue(9), Unclosed_List3):
      if limit2 <= 0:
        break
      limit2 -= 1
      print(Unclosed_List3)

  print('\nEnd fifth test\n')

  """
  Expected output
  
  LinkedList([Var()]).is_instantiated(): False
  1. A: [1, 2, 3], B: ('[]', '_46', '_47')
  2. A: [1, 2, 3], B: [1, 2, 3]
  3. C: [0, 2, 3]
  4. D3: 3, D2: 2, D: [3, 2, 3]
  5. E: [1, 2, 3], E_EoT.head(): 1, E_EoT.tail(): [2, 3]
  
  6. Unclosed_List1: ('[]', 'head', '_64')
  7. Unclosed_List2: ('[]', '_66', '_67')
  8. Unclosed_List2: [1, 2, 3]
  
  End fifth test
  """



from setuptools import setup, find_packages

def readme():
    with open('README.md') as f:
        return f.read()

setup(name='pylog',
      version='1.1',
      description='Python implementation of Prolog features.',
      long_description=readme(),
      long_description_content_type='text/markdown',
      url='https://github.com/RussAbbott/pylog',
      author='Russ Abbott, Jay Patel',
      author_email='Russ.Abbott@gmail.com, imjaypatel12@gmail.com',
      license='MIT',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      keywords='prolog python logic programming',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      include_package_data=True,
      zip_safe=False,
      install_requires=[],
      python_requires='>=3.7',
      setup_requires=["pytest-runner"],
      tests_require=["pytest"])
